@@ -157,6 +157,8 @@ static struct config_enum timestamping_enu[] = {
 	{ "hardware", TS_HARDWARE  },
 	{ "software", TS_SOFTWARE  },
 	{ "legacy",   TS_LEGACY_HW },
+	{ "onestep",  TS_ONESTEP   },
+	{ "p2p1step", TS_P2P1STEP  },
 	{ NULL, 0 },
 };
 
@@ -192,17 +194,21 @@ struct config_item config_tab[] = {
 	PORT_ITEM_INT("freq_est_interval", 1, 0, INT_MAX),
 	GLOB_ITEM_INT("gmCapable", 1, 0, 1),
 	PORT_ITEM_INT("hybrid_e2e", 0, 0, 1),
+	PORT_ITEM_INT("ignore_transport_specific", 0, 0, 1),
 	PORT_ITEM_INT("ingressLatency", 0, INT_MIN, INT_MAX),
+	GLOB_ITEM_INT("initial_delay", 0, 0, INT_MAX),
 	GLOB_ITEM_INT("kernel_leap", 1, 0, 1),
 	PORT_ITEM_INT("logAnnounceInterval", 1, INT8_MIN, INT8_MAX),
 	PORT_ITEM_INT("logMinDelayReqInterval", 0, INT8_MIN, INT8_MAX),
 	PORT_ITEM_INT("logMinPdelayReqInterval", 0, INT8_MIN, INT8_MAX),
 	PORT_ITEM_INT("logSyncInterval", 0, INT8_MIN, INT8_MAX),
 	GLOB_ITEM_INT("logging_level", LOG_INFO, PRINT_LEVEL_MIN, PRINT_LEVEL_MAX),
+	GLOB_ITEM_STR("message_tag", NULL),
 	GLOB_ITEM_STR("manufacturerIdentity", "00:00:00"),
 	GLOB_ITEM_INT("max_frequency", 900000000, 0, INT_MAX),
 	PORT_ITEM_INT("min_neighbor_prop_delay", -20000000, INT_MIN, -1),
 	PORT_ITEM_INT("neighborPropDelayThresh", 20000000, 0, INT_MAX),
+	PORT_ITEM_INT("net_sync_monitor", 0, 0, 1),
 	PORT_ITEM_ENU("network_transport", TRANS_UDP_IPV4, nw_trans_enu),
 	GLOB_ITEM_INT("ntpshm_segment", 0, INT_MIN, INT_MAX),
 	GLOB_ITEM_INT("offsetScaledLogVariance", 0xffff, 0, UINT16_MAX),
@@ -237,6 +243,7 @@ struct config_item config_tab[] = {
 	GLOB_ITEM_STR("uds_address", "/var/run/ptp4l"),
 	GLOB_ITEM_INT("use_syslog", 1, 0, 1),
 	GLOB_ITEM_STR("userDescription", ""),
+	GLOB_ITEM_INT("utc_offset", CURRENT_UTC_OFFSET, 0, INT_MAX),
 	GLOB_ITEM_INT("verbose", 0, 0, 1),
 };
 
@@ -329,6 +336,7 @@ static enum parser_result parse_section_line(char *s, enum config_section *secti
 }
 
 static enum parser_result parse_item(struct config *cfg,
+				     int commandline,
 				     const char *section,
 				     const char *option,
 				     const char *value)
@@ -387,7 +395,7 @@ static enum parser_result parse_item(struct config *cfg,
 				return NOT_PARSED;
 			}
 		}
-	} else if (cgi->flags & CFG_ITEM_LOCKED) {
+	} else if (!commandline && cgi->flags & CFG_ITEM_LOCKED) {
 		/* This global option was set on the command line. */
 		return PARSED_OK;
 	} else {
@@ -415,6 +423,10 @@ static enum parser_result parse_item(struct config *cfg,
 		dst->flags |= CFG_ITEM_DYNSTR;
 		break;
 	}
+
+	if (commandline) {
+		dst->flags |= CFG_ITEM_LOCKED;
+	}
 	return PARSED_OK;
 }
 
@@ -490,6 +502,25 @@ static void check_deprecated_options(const char **option)
 	}
 }
 
+static struct option *config_alloc_longopts(struct config *cfg)
+{
+	struct config_item *ci;
+	struct option *opts;
+	int i;
+
+	opts = calloc(1, (1 + N_CONFIG_ITEMS) * sizeof(*opts));
+	if (!opts) {
+		return NULL;
+	}
+	for (i = 0; i < N_CONFIG_ITEMS; i++) {
+		ci = &config_tab[i];
+		opts[i].name = ci->label;
+		opts[i].has_arg = required_argument;
+	}
+
+	return opts;
+}
+
 int config_read(char *name, struct config *cfg)
 {
 	enum config_section current_section = UNKNOWN_SECTION;
@@ -554,7 +585,7 @@ int config_read(char *name, struct config *cfg)
 
 		check_deprecated_options(&option);
 
-		parser_res = parse_item(cfg, current_section == GLOBAL_SECTION ?
+		parser_res = parse_item(cfg, 0, current_section == GLOBAL_SECTION ?
 					NULL : current_port->name, option, value);
 
 		switch (parser_res) {
@@ -607,7 +638,6 @@ struct interface *config_create_interface(char *name, struct config *cfg)
 	}
 
 	strncpy(iface->name, name, MAX_IFNAME_SIZE);
-	sk_get_ts_info(iface->name, &iface->ts_info);
 	STAILQ_INSERT_TAIL(&cfg->interfaces, iface, list);
 	cfg->n_interfaces++;
 
@@ -627,8 +657,15 @@ struct config *config_create(void)
 	}
 	STAILQ_INIT(&cfg->interfaces);
 
+	cfg->opts = config_alloc_longopts(cfg);
+	if (!cfg->opts) {
+		free(cfg);
+		return NULL;
+	}
+
 	cfg->htab = hash_create();
 	if (!cfg->htab) {
+		free(cfg->opts);
 		free(cfg);
 		return NULL;
 	}
@@ -657,6 +694,7 @@ struct config *config_create(void)
 	return cfg;
 fail:
 	hash_destroy(cfg->htab, NULL);
+	free(cfg->opts);
 	free(cfg);
 	return NULL;
 }
@@ -670,6 +708,7 @@ void config_destroy(struct config *cfg)
 		free(iface);
 	}
 	hash_destroy(cfg->htab, config_item_free);
+	free(cfg->opts);
 	free(cfg);
 }
 
@@ -720,6 +759,71 @@ char *config_get_string(struct config *cfg, const char *section,
 	return ci->val.s;
 }
 
+int config_harmonize_onestep(struct config *cfg)
+{
+	enum timestamp_type tstype = config_get_int(cfg, NULL, "time_stamping");
+	int two_step_flag = config_get_int(cfg, NULL, "twoStepFlag");
+
+	switch (tstype) {
+	case TS_SOFTWARE:
+	case TS_LEGACY_HW:
+		if (!two_step_flag) {
+			pr_err("one step is only possible "
+			       "with hardware time stamping");
+			return -1;
+		}
+		break;
+	case TS_HARDWARE:
+		if (!two_step_flag) {
+			pr_debug("upgrading to one step time stamping "
+				 "in order to match the twoStepFlag");
+			if (config_set_int(cfg, "time_stamping", TS_ONESTEP)) {
+				return -1;
+			}
+		}
+		break;
+	case TS_ONESTEP:
+	case TS_P2P1STEP:
+		if (two_step_flag) {
+			pr_debug("one step mode implies twoStepFlag=0, "
+				 "clearing twoStepFlag to match");
+			if (config_set_int(cfg, "twoStepFlag", 0)) {
+				return -1;
+			}
+		}
+		break;
+	}
+
+	return 0;
+}
+
+int config_parse_option(struct config *cfg, const char *opt, const char *val)
+{
+	enum parser_result result;
+
+	result = parse_item(cfg, 1, NULL, opt, val);
+
+	switch (result) {
+	case PARSED_OK:
+		return 0;
+	case NOT_PARSED:
+		fprintf(stderr, "unknown option %s\n", opt);
+		break;
+	case BAD_VALUE:
+		fprintf(stderr, "%s is a bad value for option %s\n", val, opt);
+		break;
+	case MALFORMED:
+		fprintf(stderr, "%s is a malformed value for option %s\n",
+			val, opt);
+		break;
+	case OUT_OF_RANGE:
+		fprintf(stderr, "%s is an out of range value for option %s\n",
+			val, opt);
+		break;
+	}
+	return -1;
+}
+
 int config_set_double(struct config *cfg, const char *option, double val)
 {
 	struct config_item *ci = config_find_item(cfg, NULL, option);
