@@ -127,60 +127,12 @@ static char *text2str(struct PTPText *text)
 	return (char*)(s.text);
 }
 
-#define MAX_PRINT_BYTES 16
-#define BIN_BUF_SIZE (MAX_PRINT_BYTES * 3 + 1)
-
-static char *bin2str_impl(Octet *data, int len, char *buf, int buf_len)
-{
-	int i, offset = 0;
-	if (len > MAX_PRINT_BYTES)
-		len = MAX_PRINT_BYTES;
-	buf[0] = '\0';
-	if (!data)
-		return buf;
-	if (len)
-		offset += snprintf(buf, buf_len, "%02hhx", data[0]);
-	for (i = 1; i < len; i++) {
-		if (offset >= buf_len)
-			/* truncated output */
-			break;
-		offset += snprintf(buf + offset, buf_len - offset, ":%02hhx", data[i]);
-	}
-	return buf;
-}
-
 static char *bin2str(Octet *data, int len)
 {
 	static char buf[BIN_BUF_SIZE];
 	return bin2str_impl(data, len, buf, sizeof(buf));
 }
 
-static uint16_t align16(uint16_t *p)
-{
-	uint16_t v;
-	memcpy(&v, p, sizeof(v));
-	return v;
-}
-
-static char *portaddr2str(struct PortAddress *addr)
-{
-	static char buf[BIN_BUF_SIZE];
-	switch(align16(&addr->networkProtocol)) {
-	case TRANS_UDP_IPV4:
-		if (align16(&addr->addressLength) == 4
-			&& inet_ntop(AF_INET, addr->address, buf, sizeof(buf)))
-			return buf;
-		break;
-	case TRANS_UDP_IPV6:
-		if (align16(&addr->addressLength) == 16
-			&& inet_ntop(AF_INET6, addr->address, buf, sizeof(buf)))
-			return buf;
-		break;
-	}
-	bin2str_impl(addr->address, align16(&addr->addressLength), buf, sizeof(buf));
-	return buf;
-}
-
 static void pmc_show(struct ptp_message *msg, FILE *fp)
 {
 	int action;
@@ -194,8 +146,10 @@ static void pmc_show(struct ptp_message *msg, FILE *fp)
 	struct time_status_np *tsn;
 	struct grandmaster_settings_np *gsn;
 	struct mgmt_clock_description *cd;
+	struct tlv_extra *extra;
 	struct portDS *p;
 	struct port_ds_np *pnp;
+
 	if (msg_type(msg) != MANAGEMENT) {
 		return;
 	}
@@ -209,6 +163,7 @@ static void pmc_show(struct ptp_message *msg, FILE *fp)
 	if (msg->tlv_count != 1) {
 		goto out;
 	}
+	extra = TAILQ_FIRST(&msg->tlv_list);
 	tlv = (struct TLV *) msg->management.suffix;
 	if (tlv->type == TLV_MANAGEMENT) {
 		fprintf(fp, "MANAGEMENT ");
@@ -217,6 +172,7 @@ static void pmc_show(struct ptp_message *msg, FILE *fp)
 		goto out;
 	} else {
 		fprintf(fp, "unknown-tlv ");
+		goto out;
 	}
 	mgt = (struct management_tlv *) msg->management.suffix;
 	if (mgt->length == 2 && mgt->id != TLV_NULL_MANAGEMENT) {
@@ -225,33 +181,33 @@ static void pmc_show(struct ptp_message *msg, FILE *fp)
 	}
 	switch (mgt->id) {
 	case TLV_CLOCK_DESCRIPTION:
-		cd = &msg->last_tlv.cd;
+		cd = &extra->cd;
 		fprintf(fp, "CLOCK_DESCRIPTION "
 			IFMT "clockType             0x%hx"
 			IFMT "physicalLayerProtocol %s"
 			IFMT "physicalAddress       %s"
 			IFMT "protocolAddress       %hu %s",
 			align16(cd->clockType),
-                        text2str(cd->physicalLayerProtocol),
+			text2str(cd->physicalLayerProtocol),
 			bin2str(cd->physicalAddress->address,
-                                align16(&cd->physicalAddress->length)),
+				align16(&cd->physicalAddress->length)),
 			align16(&cd->protocolAddress->networkProtocol),
 			portaddr2str(cd->protocolAddress));
 		fprintf(fp, IFMT "manufacturerId        %s"
-                        IFMT "productDescription    %s",
+			IFMT "productDescription    %s",
 			bin2str(cd->manufacturerIdentity, OUI_LEN),
 			text2str(cd->productDescription));
 		fprintf(fp, IFMT "revisionData          %s",
-                        text2str(cd->revisionData));
+			text2str(cd->revisionData));
 		fprintf(fp, IFMT "userDescription       %s"
-                        IFMT "profileId             %s",
+			IFMT "profileId             %s",
 			text2str(cd->userDescription),
 			bin2str(cd->profileIdentity, PROFILE_ID_LEN));
 		break;
 	case TLV_USER_DESCRIPTION:
 		fprintf(fp, "USER_DESCRIPTION "
 			IFMT "userDescription  %s",
-			text2str(msg->last_tlv.cd.userDescription));
+			text2str(extra->cd.userDescription));
 		break;
 	case TLV_DEFAULT_DATA_SET:
 		dds = (struct defaultDS *) mgt->data;
@@ -515,14 +471,6 @@ static void do_set_action(int action, int index, char *str)
 	}
 	switch (code) {
 	case TLV_PRIORITY1:
-		cnt = sscanf(str,  " %*s %*s %hhu", &mtd.val);
-		if (cnt != 1) {
-			fprintf(stderr, "%s SET needs 1 value\n",
-				idtab[index].name);
-			break;
-		}
-		pmc_send_set_action(pmc, code, &mtd, sizeof(mtd));
-		break;
 	case TLV_PRIORITY2:
 		cnt = sscanf(str,  " %*s %*s %hhu", &mtd.val);
 		if (cnt != 1) {
@@ -717,6 +665,7 @@ static void usage(char *progname)
 		" Other Options\n\n"
 		" -b [num]  boundary hops, default 1\n"
 		" -d [num]  domain number, default 0\n"
+		" -f [file] read configuration from 'file'\n"
 		" -h        prints this message and exits\n"
 		" -i [dev]  interface device to use, default 'eth0'\n"
 		"           for network and '/var/run/pmc.$pid' for UDS.\n"
@@ -731,13 +680,14 @@ static void usage(char *progname)
 int main(int argc, char *argv[])
 {
 	const char *iface_name = NULL;
-	char *progname;
-	int c, cnt, length, tmo = -1, batch_mode = 0, zero_datalen = 0;
+	char *config = NULL, *progname;
+	int c, cnt, index, length, tmo = -1, batch_mode = 0, zero_datalen = 0;
 	int ret = 0;
 	char line[1024], *command = NULL, uds_local[MAX_IFNAME_SIZE + 1];
 	enum transport_type transport_type = TRANS_UDP_IPV4;
 	UInteger8 boundary_hops = 1, domain_number = 0, transport_specific = 0;
 	struct ptp_message *msg;
+	struct option *opts;
 	struct config *cfg;
 #define N_FD 2
 	struct pollfd pollfd[N_FD];
@@ -749,28 +699,55 @@ int main(int argc, char *argv[])
 		return -1;
 	}
 
+	opts = config_long_options(cfg);
+
 	/* Process the command line arguments. */
 	progname = strrchr(argv[0], '/');
 	progname = progname ? 1+progname : argv[0];
-	while (EOF != (c = getopt(argc, argv, "246u""b:d:hi:s:t:vz"))) {
+	while (EOF != (c = getopt_long(argc, argv, "246u""b:d:f:hi:s:t:vz",
+				       opts, &index))) {
 		switch (c) {
+		case 0:
+			if (config_parse_option(cfg, opts[index].name, optarg)) {
+				ret = -1;
+				goto out;
+			}
+			break;
 		case '2':
-			transport_type = TRANS_IEEE_802_3;
+			if (config_set_int(cfg, "network_transport", TRANS_IEEE_802_3)) {
+				ret = -1;
+				goto out;
+			}
 			break;
 		case '4':
-			transport_type = TRANS_UDP_IPV4;
+			if (config_set_int(cfg, "network_transport", TRANS_UDP_IPV4)) {
+				ret = -1;
+				goto out;
+			}
 			break;
 		case '6':
-			transport_type = TRANS_UDP_IPV6;
+			if (config_set_int(cfg, "network_transport", TRANS_UDP_IPV6)) {
+				ret = -1;
+				goto out;
+			}
 			break;
 		case 'u':
-			transport_type = TRANS_UDS;
+			if (config_set_int(cfg, "network_transport", TRANS_UDS)) {
+				ret = -1;
+				goto out;
+			}
 			break;
 		case 'b':
 			boundary_hops = atoi(optarg);
 			break;
 		case 'd':
-			domain_number = atoi(optarg);
+			if (config_set_int(cfg, "domainNumber", atoi(optarg))) {
+				ret = -1;
+				goto out;
+			}
+			break;
+		case 'f':
+			config = optarg;
 			break;
 		case 'i':
 			iface_name = optarg;
@@ -788,8 +765,12 @@ int main(int argc, char *argv[])
 			}
 			break;
 		case 't':
-			if (1 == sscanf(optarg, "%x", &c))
-				transport_specific = c << 4;
+			if (1 == sscanf(optarg, "%x", &c)) {
+				if (config_set_int(cfg, "transportSpecific", c)) {
+					ret = -1;
+					goto out;
+				}
+			}
 			break;
 		case 'v':
 			version_show(stdout);
@@ -810,6 +791,15 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (config && (c = config_read(config, cfg))) {
+		config_destroy(cfg);
+		return -1;
+	}
+
+	transport_type = config_get_int(cfg, NULL, "network_transport");
+	transport_specific = config_get_int(cfg, NULL, "transportSpecific") << 4;
+	domain_number = config_get_int(cfg, NULL, "domainNumber");
+
 	if (!iface_name) {
 		if (transport_type == TRANS_UDS) {
 			snprintf(uds_local, sizeof(uds_local),
@@ -907,6 +897,8 @@ int main(int argc, char *argv[])
 
 	pmc_destroy(pmc);
 	msg_cleanup();
+
+out:
 	config_destroy(cfg);
 	return ret;
 }
