@@ -66,8 +66,11 @@ struct pmc *pmc_create(struct config *cfg, enum transport_type transport_type,
 		       UInteger8 domain_number, UInteger8 transport_specific,
 		       int zero_datalen)
 {
+	struct interface iface;
 	struct pmc *pmc;
 
+	memset(&iface, 0, sizeof(iface));
+
 	pmc = calloc(1, sizeof *pmc);
 	if (!pmc)
 		return NULL;
@@ -90,7 +93,13 @@ struct pmc *pmc_create(struct config *cfg, enum transport_type transport_type,
 		pr_err("failed to create transport");
 		goto failed;
 	}
-	if (transport_open(pmc->transport, iface_name,
+
+	strncpy(iface.name, iface_name, MAX_IFNAME_SIZE);
+	if (iface.ts_label[0] == '\0') {
+		strncpy(iface.ts_label, iface.name, MAX_IFNAME_SIZE);
+	}
+
+	if (transport_open(pmc->transport, &iface,
 			   &pmc->fdarray, TS_SOFTWARE)) {
 		pr_err("failed to open transport");
 		goto failed;
@@ -142,7 +151,7 @@ static struct ptp_message *pmc_message(struct pmc *pmc, uint8_t action)
 	return msg;
 }
 
-static int pmc_send(struct pmc *pmc, struct ptp_message *msg, int pdulen)
+static int pmc_send(struct pmc *pmc, struct ptp_message *msg)
 {
 	int err;
 
@@ -225,6 +234,8 @@ int pmc_send_get_action(struct pmc *pmc, int id)
 	int datalen, pdulen;
 	struct ptp_message *msg;
 	struct management_tlv *mgt;
+	struct tlv_extra *extra;
+
 	msg = pmc_message(pmc, GET);
 	if (!msg) {
 		return -1;
@@ -236,14 +247,21 @@ int pmc_send_get_action(struct pmc *pmc, int id)
 	mgt->id = id;
 	pdulen = msg->header.messageLength + sizeof(*mgt) + datalen;
 	msg->header.messageLength = pdulen;
-	msg->tlv_count = 1;
+
+	extra = tlv_extra_alloc();
+	if (!extra) {
+		pr_err("failed to allocate TLV descriptor");
+		return -ENOMEM;
+	}
+	extra->tlv = (struct TLV *) msg->management.suffix;
+	msg_tlv_attach(msg, extra);
 
 	if (id == TLV_CLOCK_DESCRIPTION && !pmc->zero_length_gets) {
 		/*
 		 * Make sure the tlv_extra pointers dereferenced in
 		 * mgt_pre_send() do point to something.
 		 */
-		struct mgmt_clock_description *cd = &msg->last_tlv.cd;
+		struct mgmt_clock_description *cd = &extra->cd;
 		uint8_t *buf = mgt->data;
 		cd->clockType = (UInteger16 *) buf;
 		buf += sizeof(*cd->clockType);
@@ -254,7 +272,7 @@ int pmc_send_get_action(struct pmc *pmc, int id)
 		cd->protocolAddress = (struct PortAddress *) buf;
 	}
 
-	pmc_send(pmc, msg, pdulen);
+	pmc_send(pmc, msg);
 	msg_put(msg);
 
 	return 0;
@@ -262,22 +280,24 @@ int pmc_send_get_action(struct pmc *pmc, int id)
 
 int pmc_send_set_action(struct pmc *pmc, int id, void *data, int datasize)
 {
-	int pdulen;
-	struct ptp_message *msg;
 	struct management_tlv *mgt;
+	struct ptp_message *msg;
+	struct tlv_extra *extra;
+
 	msg = pmc_message(pmc, SET);
 	if (!msg) {
 		return -1;
 	}
-	mgt = (struct management_tlv *) msg->management.suffix;
+	extra = msg_tlv_append(msg, sizeof(*mgt) + datasize);
+	if (!extra) {
+		return -ENOMEM;
+	}
+	mgt = (struct management_tlv *) extra->tlv;
 	mgt->type = TLV_MANAGEMENT;
 	mgt->length = 2 + datasize;
 	mgt->id = id;
 	memcpy(mgt->data, data, datasize);
-	pdulen = msg->header.messageLength + sizeof(*mgt) + datasize;
-	msg->header.messageLength = pdulen;
-	msg->tlv_count = 1;
-	pmc_send(pmc, msg, pdulen);
+	pmc_send(pmc, msg);
 	msg_put(msg);
 
 	return 0;
@@ -311,6 +331,7 @@ struct ptp_message *pmc_recv(struct pmc *pmc)
 			break;
 		case -EPROTO:
 			pr_debug("ignoring message");
+			break;
 		}
 		goto failed;
 	}
