@@ -69,10 +69,8 @@ struct foreign_clock *port_compute_best(struct port *port);
  * @param port A pointer previously obtained via port_open().
  * @param event One of the @a fsm_event codes.
  * @param mdiff Whether a new master has been selected.
- * @return Zero if the port's file descriptor array is still valid,
- *         and non-zero if it has become invalid.
  */
-int port_dispatch(struct port *p, enum fsm_event event, int mdiff);
+void port_dispatch(struct port *p, enum fsm_event event, int mdiff);
 
 /**
  * Generates state machine events based on activity on a port's file
@@ -133,13 +131,6 @@ int port_number(struct port *p);
  */
 int port_link_status_get(struct port *p);
 
-/**
- * Sets the link status for a port.
- * @param p        A port instance.
- * @param up       Pass one (1) if the link is up and zero if down.
- */
-void port_link_status_set(struct port *p, int up);
-
 /**
  * Manage a port according to a given message.
  * @param p        A pointer previously obtained via port_open().
@@ -318,9 +309,8 @@ enum fault_type last_fault_type(struct port *port);
  * @param port        A port instance.
  * @param ft          Fault type.
  * @param i           Pointer to the struct which will be filled in.
- * @return Zero on success, non-zero otherwise.
  */
-int fault_interval(struct port *port, enum fault_type ft,
-	struct fault_interval *i);
+void fault_interval(struct port *port, enum fault_type ft,
+		    struct fault_interval *i);
 
 #endif
