@@ -43,7 +43,7 @@ int sk_check_fupsync;
 
 /* private methods */
 
-static int hwts_init(int fd, const char *device, int rx_filter, int one_step)
+static int hwts_init(int fd, const char *device, int rx_filter, int tx_type)
 {
 	struct ifreq ifreq;
 	struct hwtstamp_config cfg, req;
@@ -55,7 +55,7 @@ static int hwts_init(int fd, const char *device, int rx_filter, int one_step)
 	strncpy(ifreq.ifr_name, device, sizeof(ifreq.ifr_name) - 1);
 
 	ifreq.ifr_data = (void *) &cfg;
-	cfg.tx_type    = one_step ? HWTSTAMP_TX_ONESTEP_SYNC : HWTSTAMP_TX_ON;
+	cfg.tx_type    = tx_type;
 	cfg.rx_filter  = rx_filter;
 	req = cfg;
 	err = ioctl(fd, SIOCSHWTSTAMP, &ifreq);
@@ -64,9 +64,9 @@ static int hwts_init(int fd, const char *device, int rx_filter, int one_step)
 
 	if (memcmp(&cfg, &req, sizeof(cfg))) {
 
-		pr_warning("driver changed our HWTSTAMP options");
-		pr_warning("tx_type   %d not %d", cfg.tx_type, req.tx_type);
-		pr_warning("rx_filter %d not %d", cfg.rx_filter, req.rx_filter);
+		pr_debug("driver changed our HWTSTAMP options");
+		pr_debug("tx_type   %d not %d", cfg.tx_type, req.tx_type);
+		pr_debug("rx_filter %d not %d", cfg.rx_filter, req.rx_filter);
 
 		if (cfg.tx_type != req.tx_type ||
 		    (cfg.rx_filter != HWTSTAMP_FILTER_ALL &&
@@ -159,10 +159,55 @@ failed:
 	return -1;
 }
 
+static int sk_interface_guidaddr(const char *name, unsigned char *guid)
+{
+	char file_name[64], buf[64], addr[8];
+	FILE *f;
+	char *err;
+	int res;
+
+	snprintf(file_name, sizeof buf, "/sys/class/net/%s/address", name);
+	f = fopen(file_name, "r");
+	if (!f) {
+		pr_err("failed to open %s: %m", buf);
+		return -1;
+	}
+
+	/* Set the file position to the beginning of the GUID */
+	res = fseek(f, GUID_OFFSET, SEEK_SET);
+	if (res) {
+		pr_err("fseek failed: %m");
+		goto error;
+	}
+
+	err = fgets(buf, sizeof buf, f);
+	if (err == NULL) {
+		pr_err("fseek failed: %m");
+		goto error;
+	}
+
+	res = sscanf(buf, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx:%hhx:%hhx",
+			   &addr[0], &addr[1], &addr[2], &addr[3],
+			   &addr[4], &addr[5], &addr[6], &addr[7]);
+	if (res != GUID_LEN) {
+		pr_err("sscanf failed: %m");
+		goto error;
+	}
+
+	memcpy(guid, addr, GUID_LEN);
+	fclose(f);
+
+	return 0;
+
+error:
+	fclose(f);
+	return -1;
+}
+
 int sk_interface_macaddr(const char *name, struct address *mac)
 {
 	struct ifreq ifreq;
-	int err, fd;
+	int err, fd, type;
 
 	memset(&ifreq, 0, sizeof(ifreq));
 	strncpy(ifreq.ifr_name, name, sizeof(ifreq.ifr_name) - 1);
@@ -180,11 +225,26 @@ int sk_interface_macaddr(const char *name, struct address *mac)
 		return -1;
 	}
 
-	mac->sll.sll_family = AF_PACKET;
-	mac->sll.sll_halen = MAC_LEN;
-	memcpy(mac->sll.sll_addr, &ifreq.ifr_hwaddr.sa_data, MAC_LEN);
-	mac->len = sizeof(mac->sll);
 	close(fd);
+
+	/* Get interface type */
+	type = ifreq.ifr_hwaddr.sa_family;
+	switch (type) {
+		case ARPHRD_INFINIBAND:
+			err = sk_interface_guidaddr(name, mac->sll.sll_addr);
+			if (err) {
+				pr_err("fail to get address using sysfs: %m");
+				return -1;
+			}
+			mac->sll.sll_halen = EUI64;
+			break;
+		default:
+			memcpy(mac->sll.sll_addr, &ifreq.ifr_hwaddr.sa_data, MAC_LEN);
+			mac->sll.sll_halen = EUI48;
+	}
+
+	mac->sll.sll_family = AF_PACKET;
+	mac->len = sizeof(mac->sll);
 	return 0;
 }
 
@@ -281,7 +341,7 @@ int sk_receive(int fd, void *buf, int buflen,
 				return -1;
 			}
 			sw = (struct timespec *) CMSG_DATA(cm);
-			hwts->sw = *sw;
+			hwts->sw = timespec_to_tmv(*sw);
 		}
 	}
 
@@ -295,14 +355,15 @@ int sk_receive(int fd, void *buf, int buflen,
 
 	switch (hwts->type) {
 	case TS_SOFTWARE:
-		hwts->ts = ts[0];
+		hwts->ts = timespec_to_tmv(ts[0]);
 		break;
 	case TS_HARDWARE:
 	case TS_ONESTEP:
-		hwts->ts = ts[2];
+	case TS_P2P1STEP:
+		hwts->ts = timespec_to_tmv(ts[2]);
 		break;
 	case TS_LEGACY_HW:
-		hwts->ts = ts[1];
+		hwts->ts = timespec_to_tmv(ts[1]);
 		break;
 	}
 	return cnt;
@@ -334,7 +395,7 @@ int sk_set_priority(int fd, uint8_t dscp)
 int sk_timestamping_init(int fd, const char *device, enum timestamp_type type,
 			 enum transport_type transport)
 {
-	int err, filter1, filter2 = 0, flags, one_step;
+	int err, filter1, filter2 = 0, flags, tx_type = HWTSTAMP_TX_ON;
 
 	switch (type) {
 	case TS_SOFTWARE:
@@ -344,6 +405,7 @@ int sk_timestamping_init(int fd, const char *device, enum timestamp_type type,
 		break;
 	case TS_HARDWARE:
 	case TS_ONESTEP:
+	case TS_P2P1STEP:
 		flags = SOF_TIMESTAMPING_TX_HARDWARE |
 			SOF_TIMESTAMPING_RX_HARDWARE |
 			SOF_TIMESTAMPING_RAW_HARDWARE;
@@ -359,7 +421,21 @@ int sk_timestamping_init(int fd, const char *device, enum timestamp_type type,
 
 	if (type != TS_SOFTWARE) {
 		filter1 = HWTSTAMP_FILTER_PTP_V2_EVENT;
-		one_step = type == TS_ONESTEP ? 1 : 0;
+		switch (type) {
+		case TS_SOFTWARE:
+			tx_type = HWTSTAMP_TX_OFF;
+			break;
+		case TS_HARDWARE:
+		case TS_LEGACY_HW:
+			tx_type = HWTSTAMP_TX_ON;
+			break;
+		case TS_ONESTEP:
+			tx_type = HWTSTAMP_TX_ONESTEP_SYNC;
+			break;
+		case TS_P2P1STEP:
+			tx_type = HWTSTAMP_TX_ONESTEP_P2P;
+			break;
+		}
 		switch (transport) {
 		case TRANS_UDP_IPV4:
 		case TRANS_UDP_IPV6:
@@ -374,10 +450,10 @@ int sk_timestamping_init(int fd, const char *device, enum timestamp_type type,
 		case TRANS_UDS:
 			return -1;
 		}
-		err = hwts_init(fd, device, filter1, one_step);
+		err = hwts_init(fd, device, filter1, tx_type);
 		if (err) {
 			pr_info("driver rejected most general HWTSTAMP filter");
-			err = hwts_init(fd, device, filter2, one_step);
+			err = hwts_init(fd, device, filter2, tx_type);
 			if (err) {
 				pr_err("ioctl SIOCSHWTSTAMP failed: %m");
 				return err;
