@@ -41,6 +41,8 @@
 
 #define DEFAULT_RUNDIR "/var/run/timemaster"
 
+#define DEFAULT_FIRST_SHM_SEGMENT 0
+
 #define DEFAULT_NTP_PROGRAM CHRONYD
 #define DEFAULT_NTP_MINPOLL 6
 #define DEFAULT_NTP_MAXPOLL 10
@@ -72,6 +74,7 @@ struct ntp_server {
 	int minpoll;
 	int maxpoll;
 	int iburst;
+	char *ntp_options;
 };
 
 struct ptp_domain {
@@ -81,6 +84,7 @@ struct ptp_domain {
 	double delay;
 	char **interfaces;
 	char **ptp4l_settings;
+	char *ntp_options;
 };
 
 struct source {
@@ -101,6 +105,7 @@ struct timemaster_config {
 	struct source **sources;
 	enum ntp_program ntp_program;
 	char *rundir;
+	int first_shm_segment;
 	struct program_config chronyd;
 	struct program_config ntpd;
 	struct program_config phc2sys;
@@ -223,10 +228,12 @@ static void source_destroy(struct source *source)
 	switch (source->type) {
 	case NTP_SERVER:
 		free(source->ntp.address);
+		free(source->ntp.ntp_options);
 		break;
 	case PTP_DOMAIN:
 		free_parray((void **)source->ptp.interfaces);
 		free_parray((void **)source->ptp.ptp4l_settings);
+		free(source->ptp.ntp_options);
 		break;
 	}
 	free(source);
@@ -235,7 +242,6 @@ static void source_destroy(struct source *source)
 static struct source *source_ntp_parse(char *parameter, char **settings)
 {
 	char *name, *value;
-	struct ntp_server ntp_server;
 	struct source *source;
 	int r = 0;
 
@@ -244,35 +250,38 @@ static struct source *source_ntp_parse(char *parameter, char **settings)
 		return NULL;
 	}
 
-	ntp_server.address = parameter;
-	ntp_server.minpoll = DEFAULT_NTP_MINPOLL;
-	ntp_server.maxpoll = DEFAULT_NTP_MAXPOLL;
-	ntp_server.iburst = 0;
+	source = xmalloc(sizeof(*source));
+	source->type = NTP_SERVER;
+	source->ntp.address = xstrdup(parameter);
+	source->ntp.minpoll = DEFAULT_NTP_MINPOLL;
+	source->ntp.maxpoll = DEFAULT_NTP_MAXPOLL;
+	source->ntp.iburst = 0;
+	source->ntp.ntp_options = xstrdup("");
 
 	for (; *settings; settings++) {
 		parse_setting(*settings, &name, &value);
 		if (!strcasecmp(name, "minpoll")) {
-			r = parse_int(value, &ntp_server.minpoll);
+			r = parse_int(value, &source->ntp.minpoll);
 		} else if (!strcasecmp(name, "maxpoll")) {
-			r = parse_int(value, &ntp_server.maxpoll);
+			r = parse_int(value, &source->ntp.maxpoll);
 		} else if (!strcasecmp(name, "iburst")) {
-			r = parse_bool(value, &ntp_server.iburst);
+			r = parse_bool(value, &source->ntp.iburst);
+		} else if (!strcasecmp(name, "ntp_options")) {
+			replace_string(value, &source->ntp.ntp_options);
 		} else {
 			pr_err("unknown ntp_server setting %s", name);
-			return NULL;
+			goto failed;
 		}
 		if (r) {
 			pr_err("invalid value %s for %s", value, name);
-			return NULL;
+			goto failed;
 		}
 	}
 
-	source = xmalloc(sizeof(*source));
-	source->type = NTP_SERVER;
-	source->ntp = ntp_server;
-	source->ntp.address = xstrdup(source->ntp.address);
-
 	return source;
+failed:
+	source_destroy(source);
+	return NULL;
 }
 
 static struct source *source_ptp_parse(char *parameter, char **settings)
@@ -288,6 +297,7 @@ static struct source *source_ptp_parse(char *parameter, char **settings)
 	source->ptp.phc2sys_poll = DEFAULT_PTP_PHC2SYS_POLL;
 	source->ptp.interfaces = (char **)parray_new();
 	source->ptp.ptp4l_settings = (char **)parray_new();
+	source->ptp.ntp_options = xstrdup("");
 
 	if (parse_int(parameter, &source->ptp.domain)) {
 		pr_err("invalid ptp_domain number %s", parameter);
@@ -305,6 +315,8 @@ static struct source *source_ptp_parse(char *parameter, char **settings)
 		} else if (!strcasecmp(name, "ptp4l_option")) {
 			parray_append((void ***)&source->ptp.ptp4l_settings,
 				      xstrdup(value));
+		} else if (!strcasecmp(name, "ntp_options")) {
+			replace_string(value, &source->ptp.ntp_options);
 		} else if (!strcasecmp(name, "interfaces")) {
 			parse_words(value, &source->ptp.interfaces);
 		} else {
@@ -354,6 +366,7 @@ static int parse_timemaster_settings(char **settings,
 				     struct timemaster_config *config)
 {
 	char *name, *value;
+	int r = 0;
 
 	for (; *settings; settings++) {
 		parse_setting(*settings, &name, &value);
@@ -368,10 +381,16 @@ static int parse_timemaster_settings(char **settings,
 			}
 		} else if (!strcasecmp(name, "rundir")) {
 			replace_string(value, &config->rundir);
+		} else if (!strcasecmp(name, "first_shm_segment")) {
+			r = parse_int(value, &config->first_shm_segment);
 		} else {
 			pr_err("unknown timemaster setting %s", name);
 			return 1;
 		}
+		if (r) {
+			pr_err("invalid value %s for %s", value, name);
+			return 1;
+		}
 	}
 
 	return 0;
@@ -486,6 +505,7 @@ static struct timemaster_config *config_parse(char *path)
 	config->sources = (struct source **)parray_new();
 	config->ntp_program = DEFAULT_NTP_PROGRAM;
 	config->rundir = xstrdup(DEFAULT_RUNDIR);
+	config->first_shm_segment = DEFAULT_FIRST_SHM_SEGMENT;
 
 	init_program_config(&config->chronyd, "chronyd",
 			    NULL, DEFAULT_CHRONYD_SETTINGS, NULL);
@@ -579,7 +599,8 @@ static char **get_ptp4l_command(struct program_config *config,
 }
 
 static char **get_phc2sys_command(struct program_config *config, int domain,
-				  int poll, int shm_segment, char *uds_path)
+				  int poll, int shm_segment, char *uds_path,
+				  char *message_tag)
 {
 	char **command = (char **)parray_new();
 
@@ -590,6 +611,7 @@ static char **get_phc2sys_command(struct program_config *config, int domain,
 		      xstrdup("-R"), string_newf("%.2f", poll > 0 ?
 						1.0 / (1 << poll) : 1 << -poll),
 		      xstrdup("-z"), xstrdup(uds_path),
+		      xstrdup("-t"), xstrdup(message_tag),
 		      xstrdup("-n"), string_newf("%d", domain),
 		      xstrdup("-E"), xstrdup("ntpshm"),
 		      xstrdup("-M"), string_newf("%d", shm_segment), NULL);
@@ -609,8 +631,8 @@ static char *get_refid(char *prefix, unsigned int number)
 };
 
 static void add_shm_source(int shm_segment, int poll, int dpoll, double delay,
-			   char *prefix, struct timemaster_config *config,
-			   char **ntp_config)
+			   char *ntp_options, char *prefix,
+			   struct timemaster_config *config, char **ntp_config)
 {
 	char *refid = get_refid(prefix, shm_segment);
 
@@ -618,15 +640,17 @@ static void add_shm_source(int shm_segment, int poll, int dpoll, double delay,
 	case CHRONYD:
 		string_appendf(ntp_config,
 			       "refclock SHM %d poll %d dpoll %d "
-			       "refid %s precision 1.0e-9 delay %.1e\n",
-			       shm_segment, poll, dpoll, refid, delay);
+			       "refid %s precision 1.0e-9 delay %.1e %s\n",
+			       shm_segment, poll, dpoll, refid, delay,
+			       ntp_options);
 		break;
 	case NTPD:
 		string_appendf(ntp_config,
 			       "server 127.127.28.%d minpoll %d maxpoll %d "
-			       "mode 1\n"
+			       "mode 1 %s\n"
 			       "fudge 127.127.28.%d refid %s\n",
-			       shm_segment, poll, poll, shm_segment, refid);
+			       shm_segment, poll, poll, ntp_options,
+			       shm_segment, refid);
 		break;
 	}
 
@@ -637,9 +661,9 @@ static int add_ntp_source(struct ntp_server *source, char **ntp_config)
 {
 	pr_debug("adding NTP server %s", source->address);
 
-	string_appendf(ntp_config, "server %s minpoll %d maxpoll %d%s\n",
+	string_appendf(ntp_config, "server %s minpoll %d maxpoll %d %s %s\n",
 		       source->address, source->minpoll, source->maxpoll,
-		       source->iburst ? " iburst" : "");
+		       source->iburst ? "iburst" : "", source->ntp_options);
 	return 0;
 }
 
@@ -649,14 +673,16 @@ static int add_ptp_source(struct ptp_domain *source,
 			  struct script *script)
 {
 	struct config_file *config_file;
-	char **command, *uds_path, **interfaces;
-	int i, j, num_interfaces, *phc, *phcs, hw_ts;
+	char **command, *uds_path, **interfaces, *message_tag;
+	int i, j, num_interfaces, *phc, *phcs, hw_ts, sw_ts;
 	struct sk_ts_info ts_info;
 
 	pr_debug("adding PTP domain %d", source->domain);
 
 	hw_ts = SOF_TIMESTAMPING_TX_HARDWARE | SOF_TIMESTAMPING_RX_HARDWARE |
 		SOF_TIMESTAMPING_RAW_HARDWARE;
+	sw_ts = SOF_TIMESTAMPING_TX_SOFTWARE | SOF_TIMESTAMPING_RX_SOFTWARE |
+		SOF_TIMESTAMPING_SOFTWARE;
 
 	for (num_interfaces = 0;
 	     source->interfaces[num_interfaces]; num_interfaces++)
@@ -678,9 +704,14 @@ static int add_ptp_source(struct ptp_domain *source,
 			return 1;
 		}
 
-		if (!ts_info.valid ||
-		    ((ts_info.so_timestamping & hw_ts) != hw_ts)) {
+		if (((ts_info.so_timestamping & hw_ts) != hw_ts)) {
 			pr_debug("interface %s: no PHC", source->interfaces[i]);
+			if ((ts_info.so_timestamping & sw_ts) != sw_ts) {
+				pr_err("time stamping not supported on %s",
+				       source->interfaces[i]);
+				free(phcs);
+				return 1;
+			}
 			continue;
 		}
 
@@ -727,6 +758,12 @@ static int add_ptp_source(struct ptp_domain *source,
 		uds_path = string_newf("%s/ptp4l.%d.socket",
 				       config->rundir, *shm_segment);
 
+		message_tag = string_newf("[%d", source->domain);
+		for (j = 0; interfaces[j]; j++)
+			string_appendf(&message_tag, "%s%s", j ? "+" : ":",
+				       interfaces[j]);
+		string_appendf(&message_tag, "]");
+
 		config_file = xmalloc(sizeof(*config_file));
 		config_file->path = string_newf("%s/ptp4l.%d.conf",
 						config->rundir, *shm_segment);
@@ -738,8 +775,9 @@ static int add_ptp_source(struct ptp_domain *source,
 		string_appendf(&config_file->content,
 			       "slaveOnly 1\n"
 			       "domainNumber %d\n"
-			       "uds_address %s\n",
-			       source->domain, uds_path);
+			       "uds_address %s\n"
+			       "message_tag %s\n",
+			       source->domain, uds_path, message_tag);
 
 		if (phcs[i] >= 0) {
 			/* HW time stamping */
@@ -750,7 +788,8 @@ static int add_ptp_source(struct ptp_domain *source,
 			command = get_phc2sys_command(&config->phc2sys,
 						      source->domain,
 						      source->phc2sys_poll,
-						      *shm_segment, uds_path);
+						      *shm_segment, uds_path,
+						      message_tag);
 			parray_append((void ***)&script->commands, command);
 		} else {
 			/* SW time stamping */
@@ -766,11 +805,12 @@ static int add_ptp_source(struct ptp_domain *source,
 		parray_append((void ***)&script->configs, config_file);
 
 		add_shm_source(*shm_segment, source->ntp_poll,
-			       source->phc2sys_poll, source->delay, "PTP",
-			       config, ntp_config);
+			       source->phc2sys_poll, source->delay,
+			       source->ntp_options, "PTP", config, ntp_config);
 
 		(*shm_segment)++;
 
+		free(message_tag);
 		free(uds_path);
 		free(interfaces);
 	}
@@ -865,12 +905,13 @@ static struct script *script_create(struct timemaster_config *config)
 	struct source *source, **sources;
 	struct config_file *ntp_config = NULL;
 	int **allocated_phcs = (int **)parray_new();
-	int ret = 0, shm_segment = 0;
+	int ret = 0, shm_segment;
 
 	script->configs = (struct config_file **)parray_new();
 	script->commands = (char ***)parray_new();
 
 	ntp_config = add_ntp_program(config, script);
+	shm_segment = config->first_shm_segment;
 
 	for (sources = config->sources; (source = *sources); sources++) {
 		switch (source->type) {
@@ -1012,6 +1053,14 @@ static int script_run(struct script *script)
 	pid_t pid, *pids;
 	int i, num_commands, status, ret = 0;
 
+	for (num_commands = 0; script->commands[num_commands]; num_commands++)
+		;
+
+	if (!num_commands) {
+		/* nothing to do */
+		return 0;
+	}
+
 	if (create_config_files(script->configs))
 		return 1;
 
@@ -1027,9 +1076,6 @@ static int script_run(struct script *script)
 		return 1;
 	}
 
-	for (num_commands = 0; script->commands[num_commands]; num_commands++)
-		;
-
 	pids = xcalloc(num_commands, sizeof(*pids));
 
 	for (i = 0; i < num_commands; i++) {
@@ -1042,15 +1088,25 @@ static int script_run(struct script *script)
 
 	/* wait for one of the blocked signals */
 	while (1) {
-		if (sigwaitinfo(&mask, &info) > 0)
-			break;
-		if (errno != EINTR) {
+		if (sigwaitinfo(&mask, &info) < 0) {
+			if (errno == EINTR)
+				continue;
 			pr_err("sigwaitinfo() failed: %m");
 			break;
 		}
-	}
 
-	pr_info("received signal %d", info.si_signo);
+		/*
+		 * assume only the first process (i.e. chronyd or ntpd) is
+		 * essential and continue if other processes terminate
+		 */
+		if (info.si_signo == SIGCHLD && info.si_pid != pids[0]) {
+			pr_info("process %d terminated (ignored)", info.si_pid);
+			continue;
+		}
+
+		pr_info("received signal %d", info.si_signo);
+		break;
+	}
 
 	/* kill all started processes */
 	for (i = 0; i < num_commands; i++) {
