@@ -18,17 +18,26 @@
  */
 #include <arpa/inet.h>
 #include <errno.h>
+#include <stdlib.h>
 #include <string.h>
 
 #include "port.h"
 #include "tlv.h"
 #include "msg.h"
 
+#define HTONS(x) (x) = htons(x)
+#define HTONL(x) (x) = htonl(x)
+#define NTOHS(x) (x) = ntohs(x)
+#define NTOHL(x) (x) = ntohl(x)
+
 #define TLV_LENGTH_INVALID(tlv, type) \
 	(tlv->length < sizeof(struct type) - sizeof(struct TLV))
 
 uint8_t ieee8021_id[3] = { IEEE_802_1_COMMITTEE };
 
+static TAILQ_HEAD(tlv_pool, tlv_extra) tlv_pool =
+	TAILQ_HEAD_INITIALIZER(tlv_pool);
+
 static void scaled_ns_n2h(ScaledNs *sns)
 {
 	sns->nanoseconds_msb = ntohs(sns->nanoseconds_msb);
@@ -43,7 +52,8 @@ static void scaled_ns_h2n(ScaledNs *sns)
 	sns->fractional_nanoseconds = htons(sns->fractional_nanoseconds);
 }
 
-static uint16_t flip16(uint16_t *p) {
+static uint16_t flip16(uint16_t *p)
+{
 	uint16_t v;
 	memcpy(&v, p, sizeof(v));
 	v = htons(v);
@@ -51,6 +61,24 @@ static uint16_t flip16(uint16_t *p) {
 	return v;
 }
 
+static int64_t host2net64_unaligned(int64_t *p)
+{
+	int64_t v;
+	memcpy(&v, p, sizeof(v));
+	v = host2net64(v);
+	memcpy(p, &v, sizeof(v));
+	return v;
+}
+
+static int64_t net2host64_unaligned(int64_t *p)
+{
+	int64_t v;
+	memcpy(&v, p, sizeof(v));
+	v = net2host64(v);
+	memcpy(p, &v, sizeof(v));
+	return v;
+}
+
 static int mgt_post_recv(struct management_tlv *m, uint16_t data_len,
 			 struct tlv_extra *extra)
 {
@@ -366,6 +394,111 @@ static void mgt_pre_send(struct management_tlv *m, struct tlv_extra *extra)
 	}
 }
 
+static int nsm_resp_post_recv(struct tlv_extra *extra)
+{
+	struct nsm_resp_tlv_head *head;
+	struct TLV *tlv = extra->tlv;
+	struct timePropertiesDS *tp;
+	struct PortAddress *paddr;
+	struct currentDS *cds;
+	struct parentDS *pds;
+	unsigned char *ptr;
+	uint16_t expected;
+
+	if (tlv->length < sizeof(*head) + sizeof(*extra->foot)
+	    - sizeof(head->type) - sizeof(head->length)) {
+		return -EBADMSG;
+	}
+	head = (struct nsm_resp_tlv_head *) tlv;
+	paddr = &head->parent_addr;
+	NTOHS(paddr->networkProtocol);
+	NTOHS(paddr->addressLength);
+
+	switch (paddr->networkProtocol) {
+	case TRANS_UDP_IPV4:
+		expected = 4;
+		break;
+	case TRANS_UDP_IPV6:
+		expected = 16;
+		break;
+	case TRANS_IEEE_802_3:
+		expected = 6;
+		break;
+	default:
+		return -EBADMSG;
+	}
+	if (paddr->addressLength != expected) {
+		return -EBADMSG;
+	}
+	if (tlv->length != sizeof(*head) + sizeof(*extra->foot) +
+	    paddr->addressLength - sizeof(head->type) - sizeof(head->length)) {
+		return -EBADMSG;
+	}
+
+	ptr = (unsigned char *) tlv;
+	ptr += sizeof(*head) + paddr->addressLength;
+	extra->foot = (struct nsm_resp_tlv_foot *) ptr;
+
+	pds = &extra->foot->parent;
+	cds = &extra->foot->current;
+	tp = &extra->foot->timeprop;
+
+	/*
+	 * At this point the alignment only 2 bytes worst case.
+	 * So we need to be careful with the 64 bit words.
+	 */
+	NTOHS(pds->parentPortIdentity.portNumber);
+	NTOHS(pds->observedParentOffsetScaledLogVariance);
+	NTOHL(pds->observedParentClockPhaseChangeRate);
+	NTOHS(pds->grandmasterClockQuality.offsetScaledLogVariance);
+
+	NTOHS(cds->stepsRemoved);
+	net2host64_unaligned(&cds->offsetFromMaster);
+	net2host64_unaligned(&cds->meanPathDelay);
+
+	NTOHS(tp->currentUtcOffset);
+
+	NTOHL(extra->foot->lastsync.seconds_lsb);
+	NTOHS(extra->foot->lastsync.seconds_msb);
+	NTOHL(extra->foot->lastsync.nanoseconds);
+
+	return 0;
+}
+
+static void nsm_resp_pre_send(struct tlv_extra *extra)
+{
+	struct nsm_resp_tlv_head *head;
+	struct timePropertiesDS *tp;
+	struct PortAddress *paddr;
+	struct currentDS *cds;
+	struct parentDS *pds;
+
+	head = (struct nsm_resp_tlv_head *) extra->tlv;
+	paddr = &head->parent_addr;
+
+	pds = &extra->foot->parent;
+	cds = &extra->foot->current;
+	tp = &extra->foot->timeprop;
+
+	NTOHS(paddr->networkProtocol);
+	NTOHS(paddr->addressLength);
+
+	HTONS(pds->parentPortIdentity.portNumber);
+	HTONS(pds->observedParentOffsetScaledLogVariance);
+	HTONL(pds->observedParentClockPhaseChangeRate);
+	HTONS(pds->grandmasterClockQuality.offsetScaledLogVariance);
+
+	HTONS(cds->stepsRemoved);
+	host2net64_unaligned(&cds->offsetFromMaster);
+	host2net64_unaligned(&cds->meanPathDelay);
+
+	HTONS(tp->currentUtcOffset);
+
+	HTONL(extra->foot->lastsync.seconds_lsb);
+	HTONS(extra->foot->lastsync.seconds_msb);
+	HTONL(extra->foot->lastsync.nanoseconds);
+}
+
 static int org_post_recv(struct organization_tlv *org)
 {
 	struct follow_up_info_tlv *f;
@@ -411,15 +544,41 @@ static void org_pre_send(struct organization_tlv *org)
 	}
 }
 
-int tlv_post_recv(struct TLV *tlv, struct tlv_extra *extra)
+struct tlv_extra *tlv_extra_alloc(void)
+{
+	struct tlv_extra *extra = TAILQ_FIRST(&tlv_pool);
+
+	if (extra) {
+		TAILQ_REMOVE(&tlv_pool, extra, list);
+	} else {
+		extra = calloc(1, sizeof(*extra));
+	}
+	return extra;
+}
+
+void tlv_extra_cleanup(void)
+{
+	struct tlv_extra *extra;
+
+	while ((extra = TAILQ_FIRST(&tlv_pool)) != NULL) {
+		TAILQ_REMOVE(&tlv_pool, extra, list);
+		free(extra);
+	}
+}
+
+void tlv_extra_recycle(struct tlv_extra *extra)
+{
+	memset(extra, 0, sizeof(*extra));
+	TAILQ_INSERT_HEAD(&tlv_pool, extra, list);
+}
+
+int tlv_post_recv(struct tlv_extra *extra)
 {
 	int result = 0;
 	struct management_tlv *mgt;
 	struct management_error_status *mes;
+	struct TLV *tlv = extra->tlv;
 	struct path_trace_tlv *ptt;
-	struct tlv_extra dummy_extra;
-	if (!extra)
-		extra = &dummy_extra;
 
 	switch (tlv->type) {
 	case TLV_MANAGEMENT:
@@ -458,6 +617,11 @@ int tlv_post_recv(struct TLV *tlv, struct tlv_extra *extra)
 	case TLV_AUTHENTICATION_CHALLENGE:
 	case TLV_SECURITY_ASSOCIATION_UPDATE:
 	case TLV_CUM_FREQ_SCALE_FACTOR_OFFSET:
+	case TLV_PTPMON_REQ:
+		break;
+	case TLV_PTPMON_RESP:
+		result = nsm_resp_post_recv(extra);
+		break;
 	default:
 		break;
 	}
@@ -496,6 +660,11 @@ void tlv_pre_send(struct TLV *tlv, struct tlv_extra *extra)
 	case TLV_AUTHENTICATION_CHALLENGE:
 	case TLV_SECURITY_ASSOCIATION_UPDATE:
 	case TLV_CUM_FREQ_SCALE_FACTOR_OFFSET:
+	case TLV_PTPMON_REQ:
+		break;
+	case TLV_PTPMON_RESP:
+		nsm_resp_pre_send(extra);
+		break;
 	default:
 		break;
 	}
